#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mylib.h"

int BFInterpret(char *Script, size_t ScriptSize);

int
main(int argc, char *argv[])
{
  char   *Script;
  size_t  ScriptSize;

  if(argc != 2)
  {
    puts("usage: bfi <ScriptName>");
    return 1;
  }

  Script = readfile(argv[1], &ScriptSize);
  if(!Script)
  {
    return 2;
  }

  BFInterpret(Script, ScriptSize);
    


  return 0;
}

int
BFInterpret(char *Script, size_t ScriptSize)
{
  static unsigned char Memory[30000];
  int Ptr = 0;
  int IP  = 0; // Instruction Pointer

  for(;;)
  {
    // End of script?
    if(IP >= (int)ScriptSize)
      break;

    switch(Script[IP])
    {
      case '>': 
        Ptr++; 
        if(Ptr == 30000) Ptr = 0;
        IP++;
        break;

       case '<':
        Ptr--;
        if(Ptr == -1) Ptr = 29999;
        IP++;
        break;

       case '+':
        Memory[Ptr]++;
        IP++;
        break;

       case '-':
        Memory[Ptr]--;
        IP++;
        break;

       case '.':
        putchar((int)Memory[Ptr]);
        IP++;
        break;

       case ',':
        Memory[Ptr] = (unsigned char)getchar();
        IP++;
        break;

       case '[':
        if(Memory[Ptr] != 0)
        {
          IP++;
        }
        else
        {
          int Depth = 0;
          int i = IP + 1;
          while(i < (int)ScriptSize)
          {
            if(Script[i] == '[') Depth++;
            else if(Script[i] == ']') Depth--;

            i++;

            if(Depth == -1)
              break;
          }

          if(Depth != -1)
          {
            printf("error: missing pair ] at %i\n", IP);
            return 1;
          }

          IP = i;
        }
        break;

       case ']':
       {
         int Depth = 0;
         int i = IP - 1;
         while(i > 0)
         {
           if(Script[i] == '[') Depth++;
           else if(Script[i] == ']') Depth--;

           i--;

           if(Depth == 1)
             break;
         }

         if(Depth != 1)
         {
           printf("error: missing pair [ at %i\n", IP);
           return 2;
         }
         IP = i;
        }
       break;

       default:
       IP++;
    }
    
  }
   

  // All OK
  return 0;
}

